{$R- range checking}
{$C- i/o breakable}

program g_wiz;

procedure Graphics;                                   external 'GRAPH.BIN';
procedure HiRes;                                      external Graphics[6];
procedure ClearScreen;                                external Graphics[60];
procedure Plot(X,Y,Color: Integer);                   external Graphics[21];
procedure Draw(X1,Y1,X2,Y2,Color: Integer);           external Graphics[24];
procedure Circle(X,Y,Radius,Color: Integer);          external Graphics[33];

{ This is a gravitation workshop program, the user can enter many circular
  bodies, specify their initial positions and velocities, and let them go.
  Run the program and press ? for help.  }

const
  max_bodies = 100;
  white = 1;
  black = 0;
  ox = 151;
  oy = 99;
  maxstep = 8;
  step1 = 1;
  int_limit = 32000;

type
  body_t = record
    mass,
    radius,
    dx,dy,x,y : real;     { dx,dy are x,y velocities }
  end;

var         { GLOBAL VARIABLES }

  body : array[1..max_bodies]of body_t;
  dist : array[1..max_bodies,1..max_bodies]of real;
  num_bodies : integer;
  trace : boolean;
  g : real;          { gravity const }
  dir : integer;     { clock-wise or counter }

  { how the variables are used, (this comment does not necessarily reflect this
    version of the program): body is an array of all the bodies on the screen.
    If one gets merged then the last body get copied onto it and num_bodies is
    decremented.  Dist is the distances array, for all 0 < i < j <= num_bodies,
    dist[i,j] is the distance from i to j, but dist[j,i] is the square of the
    distance. }


procedure helpscreen;

  var ch:char;

  begin
    textmode(bw40);
    writeln(#213,#205,#205,#205,#205,#205,#205,#205,#205,#205,#205,#205,#205,#205,
            ' G - W I Z ',
            #205,#205,#205,#205,#205,#205,#205,#205,#205,#205,#205,#205,#205,#184);
    writeln(#179,'        gravitation workshop         ',#179);
    writeln(#179,'     by john henckel, july 1986      ',#179);
    writeln(#179,'              freeware               ',#179);
    writeln(#179,'press:                               ',#179);
    writeln(#179,'  space=start/stop,  esc  =quit      ',#179);
    writeln(#179,'  bkspc=prev body,   retrn=next body ',#179);
    writeln(#179,'  ins  =new body,    del  =delete    ',#179);
    writeln(#179,'  ',#27,#26,#24,#25,
                       ' =move body, w/shift=velocity  ',#179);
    writeln(#179,'  f7-f8=radius,     f9-f10=mass      ',#179);
    writeln(#179,'  t =trace toggl,     r =reset body  ',#179);
    writeln(#179,'  o =orbit 2 bodies,  z =0 velocity  ',#179);
    writeln(#179,'  m =adjust body to 0 tot. momentum  ',#179);
    writeln(#179,'  n =adjust other bodies to 0 tot.m  ',#179);
    writeln(#179,'  e =detail editor.                  ',#179);
    writeln(#179,'                                     ',#179);
    writeln(#179,'  press any key to continue.         ',#179);
    writeln(#212,#205,#205,#205,#205,#205,#205,#205,#205,#205,
            #205,#205,#205,#205,#205,#205,#205,#205,#205,#205,
            #205,#205,#205,#205,#205,#205,#205,#205,#205,#205,
            #205,#205,#205,#205,#205,#205,#205,#205,#190);
    read(kbd,ch);
    graphics;
  end;


function rounded(r:real):real;

  begin
    rounded:=round(r*step1)/step1;
  end;


procedure burp;

  begin
    sound(50); delay(150); nosound;
  end;


procedure drawbody(x,y,r:real; color:integer);

  begin
    if (abs(x)<int_limit) and (abs(y)<int_limit) then
      if r<0.5 then
        plot(round(x),round(y),color)
      else
        circle(round(x),round(y),round(r),color);
  end;


procedure drawbodies;

  var n:integer;

  begin
    for n:=1 to num_bodies do
      with body[n] do
        drawbody(ox+x,oy-y,radius,white);
  end;


procedure drawvectors;

  var n:integer;

  begin
    for n:=1 to num_bodies do
      with body[n] do
        if (abs(x)<int_limit) and (abs(y)<int_limit) then
          draw(round(ox+x),round(oy-y),round(ox+x+dx),round(oy-(y+dy)),white);
  end;


procedure mergebodies(a,b:integer);

  { merges body b into body a }

  var mm:real;

  begin
    if not trace then
      drawbody(ox+body[b].x,oy-body[b].y,body[b].radius,black);
    with body[a] do begin
      if not trace then drawbody(ox+x,oy-y,radius,black);
      mm:=mass+body[b].mass;
      dx:=(mass*dx+body[b].mass*body[b].dx)/mm;
      dy:=(mass*dy+body[b].mass*body[b].dy)/mm;
      radius:=sqrt(sqr(radius)+sqr(body[b].radius));
      mass:=mm;
      drawbody(ox+x,oy-y,radius,white);
    end;
    body[b]:=body[num_bodies];
    if num_bodies>1 then num_bodies:=num_bodies-1;
  end;


procedure compute_vel;

  { This computes the new dx and dy values for body-n, ajusting for the
    gravitational pull of all the other bodies.  }

  var n,i:integer;
      force:real;

  begin
    for n:=1 to num_bodies do with body[n] do begin

      { compute gravity effect }

      for i:=1 to (n-1) do begin
        force:=g*body[i].mass/dist[n,i];
        dx:=dx+(force*(body[i].x-x)/dist[i,n]);
        dy:=dy+(force*(body[i].y-y)/dist[i,n]);
      end;
      for i:=(n+1) to num_bodies do begin
        force:=g*body[i].mass/dist[i,n];
        dx:=dx+(force*(body[i].x-x)/dist[n,i]);
        dy:=dy+(force*(body[i].y-y)/dist[n,i]);
      end;
    end;
  end;


procedure adjust_dist;

  { This moves the body and checks if the body crashed into any other body.  }

  label bye;
  const thresh = 0.0001;
  var n,i:integer;

  begin
    n:=1;
    repeat with body[n] do begin
      if not trace then drawbody(ox+x,oy-y,radius,black);
      x:=x+dx;
      y:=y+dy;
      drawbody(ox+x,oy-y,radius,white);
      for i:=1 to (n-1) do begin
        dist[n,i]:=sqr(body[i].x-x)+sqr(body[i].y-y);
        if abs(dist[n,i])<thresh then dist[n,i]:=thresh;
        dist[i,n]:=sqrt(dist[n,i]);
        if dist[i,n]<(body[i].radius+radius) then begin
          mergebodies(i,n);
          goto bye;
        end;
      end;
    end;
    n:=n+1;
  bye:
    until n>num_bodies;
  end;


procedure eqmomentum(n:integer);

  var i:integer;
      mx,my:real;

  begin
    mx:=0;  my:=0;
    for i:=1 to num_bodies do
      if i<>n then begin
        mx:=body[i].mass*body[i].dx+mx;
        my:=body[i].mass*body[i].dy+my;
      end;
    if abs(body[n].mass)<0.0000001 then body[n].mass:=0.0000001;
    body[n].dx:=-mx/body[n].mass;
    body[n].dy:=-my/body[n].mass;
  end;


procedure eqmomentum2(n:integer);

  var i:integer;
      mass:real;

  begin
    mass:=0;
    for i:=1 to num_bodies do
      if i<>n then
        mass:=mass+body[i].mass;
    if mass<>0 then
      for i:=1 to num_bodies do
        if i<>n then begin
          body[i].dx:=-body[n].mass*body[n].dx/mass;
          body[i].dy:=-body[n].mass*body[n].dy/mass;
        end;
  end;


procedure terminate;

  begin
    textmode(bw80);
    halt;
  end;


procedure gotoitem(n:integer);

  begin
    case n of
      1:gotoxy(15,5);
      2:gotoxy(15,6);
      3:gotoxy(15,9);
      4:gotoxy(15,10);
      5:gotoxy(15,13);
      6:gotoxy(15,14);
      7:gotoxy(15,17);
      8:gotoxy(15,18);
      9:gotoxy(15,19);
    end
  end;


procedure readf(var f:real; ch:char; item:integer);

  var s:string[20];
      x:real;
      p:integer;

  begin
    write(ch);
    read(s);
    val(ch+s+' ',x,p);
    if p>1 then
      f:=x;
    gotoitem(item);
    clreol;
    write(f:5:7);
    gotoitem(item);
  end;


procedure edit_detail(n:integer);

  label top;
  const num_items = 8;
  var ch:char;
      quit:boolean;
      item:integer;

  begin
    textmode(bw40);
    quit:=false;
 top:
    clrscr;
    with body[n] do begin
      writeln('detail editor');
      writeln;
      writeln('body# ',n);
      writeln;
      writeln('  mass      = ',mass:5:7);
      writeln('  radius    = ',radius:5:7);
      writeln;
      writeln('position ');
      writeln('          x = ',x:5:7);
      writeln('          y = ',y:5:7);
      writeln;
      writeln('velocity ');
      writeln('          x = ',dx:5:7);
      writeln('          y = ',dy:5:7);
      writeln;
      writeln('universals ');
      writeln('  gravity   = ',g:5:7);
      writeln('  clockwise = ',dir);
      writeln;
      writeln;
      writeln('press ',#27,#26,' to change body number,');
      writeln('      ',#25,#24,' to move the cursor,');
      writeln('      space to exit.');

      item:=1;
      gotoitem(item);
      repeat
        read(kbd,ch);
        if ch=#27 then
          if keypressed then begin
            read(kbd,ch);
            case ch of
              #75:begin
                if n=1 then n:=num_bodies
                       else n:=n-1;
                goto top;
              end;
              #77:begin
                if n=num_bodies then n:=1
                       else n:=n+1;
                goto top;
              end;
              #72:if item=1 then item:=num_items else item:=item-1;
              #80:if item=num_items then item:=1 else item:=item+1;
              #0..#255:burp;
            end;
          end
          else
            terminate
        else
          if ch=#13 then if item=num_items then item:=1 else item:=item+1
          else if ch=' ' then quit:=true
          else begin
            clreol;
            case item of
              1:readf(mass,ch,item);
              2:if ch<>'-' then readf(radius,ch,item);
              3:readf(x,ch,item);
              4:readf(y,ch,item);
              5:readf(dx,ch,item);
              6:readf(dy,ch,item);
              7:readf(g,ch,item);
              8:begin
                  dir:=dir+1;
                  if dir>1 then dir:=-1;
                  write(dir,' ');
                  item:=item-1;
                end;
            end;
            if item=num_items then item:=1 else item:=item+1;
          end;
        gotoitem(item);
      until quit;
    end;
    graphics;
  end;


procedure orbitbody(i,j:integer);

  { this adds velocity to bodies i,j so that (if nothing diverts them)
    they will orbit each other in circles. the formula for this is that
    centrifugal-force = gravity force, thus

            V1 = M2 * sqrt( g / ( dist * (M1 + M2) ) ).

    This ADDS to the dx,dy's so they should be first set to zero for this to
    work on two body systems.  }

  var vel,d:real;
      temp:integer;

  begin
    if (j>0) and (j<=num_bodies) then begin
      d:=sqrt(sqr(body[i].x-body[j].x)+sqr(body[i].y-body[j].y));
      if abs(d)<0.0001 then d:=0.0001;
      if dir=0 then
        temp:=1-2*random(2)   { pick clockwise/counter at random }
      else
        temp:=dir;

      { compute vel for body i }

      vel:=body[j].mass*sqrt(abs(g/(d*(body[i].mass+body[j].mass))));
      body[i].dx:=body[i].dx+temp*(body[i].y-body[j].y)*vel/d;
      body[i].dy:=body[i].dy+temp*(body[j].x-body[i].x)*vel/d;

      { compute vel for body j }

      vel:=body[i].mass*sqrt(abs(g/(d*(body[j].mass+body[i].mass))));
      body[j].dx:=body[j].dx+temp*(body[j].y-body[i].y)*vel/d;
      body[j].dy:=body[j].dy+temp*(body[i].x-body[j].x)*vel/d;
    end;
  end;


procedure editor;

var step:real;
    pc,ch:char;
    n,m:integer;
    quit:boolean;

  begin
    step:=step1;
    n:=1;
    quit:=false;
    repeat             { editing loop }
      clearscreen;
      gotoxy(1,25);
      write('N:',n,'   X:',body[n].x:1:1,'  Y:',body[n].y:1:1,'  Mass:',body[n].mass:1:1,'  ');
      drawbodies;
      drawvectors;
      while not keypressed do begin       { flicker the current body }
        delay(7);
        drawbody(ox+body[n].x,oy-body[n].y,body[n].radius,black);
        drawbody(ox+body[n].x,oy-body[n].y,body[n].radius,white);
      end;
      read(kbd,ch);
      if ch=#27 then
        if keypressed then begin          { keypad and function keys }
          read(kbd,ch);
          if (pc=ch) and keypressed then begin
            if (step<maxstep) then step:=step*2
          end
          else
            step:=step1;
          case ord(ch) of
            $48:body[n].y:=body[n].y+step;
            $4b:body[n].x:=body[n].x-step;
            $4d:body[n].x:=body[n].x+step;
            $50:body[n].y:=body[n].y-step;
            $52:if num_bodies<max_bodies then begin   {insert}
                  num_bodies:=num_bodies+1;
                  body[num_bodies]:=body[n];
                  n:=num_bodies;
                  body[n].x:=random(2*ox)-ox;
                  body[n].y:=random(2*oy)-oy;
                end;
            $53:if num_bodies>1 then begin            {delete}
                  body[n]:=body[num_bodies];
                  num_bodies:=num_bodies-1;
                  if n>1 then n:=n-1;
                end;
     {F7}   $41:if body[n].radius>=step then body[n].radius:=body[n].radius-step;
            $42:body[n].radius:=body[n].radius+step;
            $43:if body[n].mass>=step then body[n].mass:=body[n].mass-step;
            $44:body[n].mass:=body[n].mass+step;
            0..255:helpscreen;
          end;
        end
        else                            { esc was pressed }
          terminate
      else begin                        { non-esc keys }
        if (pc=ch) and keypressed then begin
          if (step<maxstep) then step:=step*2
        end
        else
          step:=step1;
        case upcase(ch) of
          '8':body[n].dy:=rounded(body[n].dy)+step;
          '4':body[n].dx:=rounded(body[n].dx)-step;
          '6':body[n].dx:=rounded(body[n].dx)+step;
          '2':body[n].dy:=rounded(body[n].dy)-step;
          ' ':quit:=true;
          'E':edit_detail(n);
          'O':begin
                gotoxy(1,25);
                write('what body number?                      ');
                gotoxy(19,25);
         {$I-}  read(m);         {$I+}
                orbitbody(n,m);
                m:=ioresult;
              end;
          'M':eqmomentum(n);
          'N':eqmomentum2(n);
          'T':trace:=not trace;
          'R':with body[n] do begin
                x:=random(2*ox)-ox;
                y:=random(2*oy)-oy;
                dx:=0;
                dy:=0;
              end;
          'Z':with body[n] do begin
                dx:=0;
                dy:=0;
              end;
          #13:if n=num_bodies then n:=1 else n:=n+1;
          #8:if n=1 then n:=num_bodies else n:=n-1;
          #0..#255:helpscreen;
        end;
      end;
      pc:=ch;   { previous char }
    until quit;
    clearscreen;
    drawbodies;
  end;          { editor }


procedure readkeys;

  { This iterprets the action-time key presses. }

  var ch:char;

  begin
    read(kbd,ch);
    case upcase(ch) of
      #27:if keypressed then editor
                        else terminate;
      'T':trace:=not trace;
      #0..#255:editor;
    end;
  end;


begin

  { make the first body, (there can never be fewer than one) }

  num_bodies:=1;
  body[1].mass:=9;
  body[1].radius:=3;
  body[1].dx:=0;
  body[1].dy:=0;
  body[1].x:=0;
  body[1].y:=0;

  trace:=false;
  dir:=1;
  g:=9.8;
  randomize;

  graphics;
  editor;
  repeat                               { action loop }
    if keypressed then readkeys;
    adjust_dist;                       { adjust position and recompute dist }
    compute_vel;                       { compute new dx and dy's }
  until false;
end.